# -*- coding: utf-8 -*-
"""backup module"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import json
import logging
import time

from datetime import datetime
from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.modules import DWhoModuleBase, MODULES
from dwho.config import DWHO_SHARED
from sonicprobe.libs import xys
from sonicprobe.libs.http_json_server import HttpReqErrJson
from sonicprobe.libs.moresynchro import RWLock


LOG        = logging.getLogger('xtrargon.modules.backup')
_MONIT_KEY = "supervision:dashboardv2:backup:"
_STATUS    = {'processing': 'IN PROGRESS',
              'successful': 'OK',
              'failed':     'KO'}


class XtrargonBackupModule(DWhoModuleBase):
    MODULE_NAME = 'backup'

    LOCK        = RWLock()

    def safe_init(self, options):
        self.server_id      = self.config['general']['server_id']
        self.lock_timeout   = self.config['general']['lock_timeout']
        self.adapter_redis  = DWhoAdapterRedis(self.config, prefix = 'supervision')
        self.server_key     = "X%s" % self.server_id.split('-')[-1].rstrip('ab')
        self.monit_key      = _MONIT_KEY + self.server_key
        DWHO_SHARED.add('xtrargon')

        if not self.adapter_redis.servers:
            raise SystemError("missing redis server")


    NOTIFY_SCHEMA = xys.load("""
    backup!-: !!str
    type!-: !!str
    full*-: !!str
    status!-: !~~enum(processing,failed,successful)
    backup_started*-: !!str
    backup_ended*-: !!str
    prepare_started*-: !!str
    prepare_ended*-: !!str
    compressed*-: !!bool False
    compressed_with*-: !!str
    binary_log_name*-: !!str
    master_binary_log_name*-: !!str
    hostname!-: !!str
    message*-: !!str
    """)

    def notify(self, request):
        args = request.payload_params()

        if not xys.validate(args, self.NOTIFY_SCHEMA):
            raise HttpReqErrJson(415, "invalid arguments for command")

        if not self.LOCK.acquire_write(self.lock_timeout):
            raise HttpReqErrJson(503, "unable to take LOCK for writing after %s seconds" % self.lock_timeout)

        try:
            obj     = {'title': "%s - %s" % (self.server_key, "xtrargon: %s" % self.server_id),
                       'status': 'WARNING',
                       'kpis': None,
                       'details': {},
                       'timestamp': int(time.time())}

            if args['type'] not in ('full', 'incr'):
                return

            data    = self.adapter_redis.get_key(self.monit_key)

            if data:
                data = json.loads(data)
            else:
                data = obj

            data['kpis']   = None
            data['details'][self.server_id] = args
            data['status'] = _STATUS.get(args['status'], 'WARNING')

            if args.get('backup_started') and args.get('backup_ended'):
                if args['backup_started'].count('_') == 4:
                    xformat = '%Y_%m_%d-%H_%M_%S'
                else:
                    xformat = '%Y-%m-%d_%H-%M-%S'

                started      = datetime.strptime(args['backup_started'], xformat)
                ended        = datetime.strptime(args['backup_ended'], xformat)
                data['kpis'] = (ended - started).total_seconds()

            DWHO_SHARED.set('xtrargon', 'state', args)

            LOG.debug("set key: %r with data: %r", self.monit_key, data)

            self.adapter_redis.set_key(self.monit_key, json.dumps(data))

            return data
        except HttpReqErrJson:
            raise
        except Exception, e:
            LOG.exception(e)
            raise HttpReqErrJson(503, repr(e))
        finally:
            self.LOCK.release()


if __name__ != "__main__":
    def _start():
        MODULES.register(XtrargonBackupModule())
    _start()
