# -*- coding: utf-8 -*-
"""stream plugin"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import bitmath
import boto3
import botocore
import cPickle as pickle
import gc
import logging
import os
import Queue
import resource
import threading
import time
import zlib

from dwho.classes.plugins import DWhoPluginBase, PLUGINS
from sonicprobe import helpers
from sonicprobe.libs import keystore, threading_tcp_server, workerpool, xbstream
from SocketServer import StreamRequestHandler


LOG = logging.getLogger('xtrargon.plugins.stream')

logging.getLogger('boto3').setLevel(logging.WARNING)
logging.getLogger('botocore').setLevel(logging.WARNING)
logging.getLogger('requests').setLevel(logging.WARNING)
logging.getLogger('xbstream').setLevel(logging.WARNING)

DEFAULT_TIMEOUT         = 60
DEFAULT_COMPRESS_LEVEL  = 6
DEFAULT_MAX_MEMORY      = '1G'

S3_MIN_CHUNK_SIZE       = (5 * 1024 * 1024)
S3_MAX_CHUNK_SIZE       = (20 * 1024 * 1024)
OBJECT_MAX_LIFETIME     = 216000
WORKER_LIFETIME         = 300
JOB_MAX_QUEUE_SIZE      = 25
RLIMIT_NOFILE           = 65536
_KILLED                 = False
_KEYSTORE               = keystore.Keystore()
_KEYUID                 = keystore.Keystore()
_SYS                    = keystore.Keystore()
_QUPJOB                 = Queue.Queue()


class XtrargonWriteFile(object):
    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object writefile: doesn't exist: %r", key)
            return

        (body_size, payload, fo) = (0, None, None)
        try:
            if _KEYSTORE.get(key, 'size') == 0 \
               and _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                return

            dirname  = _KEYSTORE.get(key, 'dirname')
            filename = _KEYSTORE.get(key, 'filename')
            fullpath = os.path.join(dirname, filename)
            helpers.make_dirs(os.path.dirname(fullpath))

            if not _KEYSTORE.get(key, 'fopen'):
                fd = os.open(fullpath, os.O_RDWR|os.O_CREAT)
                _KEYSTORE.set(key, 'fopen', os.fdopen(fd, 'wb+'))

            fo        = _KEYSTORE.get(key, 'fopen')

            if _KEYSTORE.get(key, 'size') == 0 \
               and _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF:
                if fo:
                    fo.close()
                _KEYSTORE.delete(key)
                return

            size      = _KEYSTORE.get(key, 'size')
            payload   = _KEYSTORE.get(key, 'payload')
            wait_part = _KEYSTORE.get(key, 'wait_part') or 1

            LOG.info("object writef: %r, size: %r", key, size)

            for no in range(wait_part, len(payload) + 2):
                if no not in payload:
                    _KEYSTORE.set(key, 'wait_part', no)
                    break
                (s, b)      = payload.pop(no)
                if s is not None:
                    body_size  += s
                    fo.write(b)
                (s, b)      = (None, None)
                payload[no] = (s, b)

            _KEYSTORE.set(key, 'size', size - body_size)

            if _KEYSTORE.get(key, 'size') == 0 \
               and _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF:
                if fo:
                    fo.close()
                _KEYSTORE.delete(key)

            LOG.info("object writef wrote: %r, size: %r", key, size)
        except Exception, e:
            if fo:
                fo.close()
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            del payload
            gc.collect()


class XtrargonUploadNormal(object):
    def __init__(self, fsdbucket, fsdclient):
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient

    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object normal: doesn't exist: %r", key)
            return

        (payload, x) = (None, None)

        try:
            if _KEYSTORE.get(key, 'size') == 0 \
               or _KEYSTORE.get(key, 'id'):
                return

            payload = _KEYSTORE.get(key, 'payload')

            size    = _KEYSTORE.get(key, 'size')
            _KEYSTORE.set(key, 'payload', {})
            _KEYSTORE.delete(key)

            LOG.info("put object normal: %r, size: %r", key, size)

            self.fsdclient.put_object(Bucket = self.fsdbucket,
                                      Key    = key,
                                      Body   = ''.join(x[1] for x in payload.itervalues()))

            LOG.info("object normal transferred: %r, size: %r", key, size)
        except Exception, e:
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            del payload, x
            gc.collect()


class XtrargonUploadPart(object):
    def __init__(self, fsdbucket, fsdclient):
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient

    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object part: doesn't exist: %r", key)
            return

        (body, body_size, payload) = ("", 0, None)

        try:
            if _KEYSTORE.get(key, 'size') < S3_MIN_CHUNK_SIZE \
               and _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                return

            payload    = _KEYSTORE.get(key, 'payload')
            if not payload:
                return

            wait_part  = _KEYSTORE.get(key, 'wait_part') or 1
            start_part = _KEYSTORE.get(key, 'start_part') or wait_part
            _KEYSTORE.set(key, 'start_part', None)

            for no in range(start_part, len(payload) + 2):
                if no not in payload:
                    _KEYSTORE.set(key, 'wait_part', no)
                    break
                (s, b)      = payload.pop(no)
                if s is not None:
                    body_size  += s
                    body       += b
                (s, b)      = (None, None)
                payload[no] = (s, b)

                if body_size >= S3_MIN_CHUNK_SIZE:
                    break

            if body_size == 0:
                return

            if body_size < S3_MIN_CHUNK_SIZE \
               and _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                _KEYSTORE.set(key, 'start_part', no - 1)
                payload[no - 1] = (body_size, body)
                return

            part_number = _KEYSTORE.get(key, 'part_number') + 1
            upload_id   = _KEYSTORE.get(key, 'id')
            size        = _KEYSTORE.get(key, 'size')

            if upload_id is True:
                m = self.fsdclient.create_multipart_upload(
                        Bucket = self.fsdbucket,
                        Key    = key)
                upload_id = m['UploadId']
                _KEYSTORE.set(key, 'id', m['UploadId'])

            (_KEYSTORE.set(key, 'size', size - body_size)
                      .set(key, 'part_number', part_number)
                      .try_release(key))

            LOG.info("put object part: %r, part: %r, body_size: %r, size: %r",
                     key,
                     part_number,
                     body_size,
                     size)

            part = self.fsdclient.upload_part(Bucket     = self.fsdbucket,
                                              Key        = key,
                                              Body       = body,
                                              UploadId   = upload_id,
                                              PartNumber = part_number)

            LOG.info("object part transferred: %r, part: %r, size: %r", key, part_number, body_size)

            if not _KEYSTORE.acquire(key, exists = True):
                LOG.error("object part process: object doesn't exist: %r", key)
                return

            parts = _KEYSTORE.get(key, 'parts')
            parts.append({'PartNumber': part_number,
                          'ETag':       part['ETag']})

            _KEYSTORE.set(key, 'parts', parts)

            if _KEYSTORE.get(key, 'size') == 0 \
               and _KEYSTORE.get(key, 'last_chunk_type') == xbstream.XB_CHUNK_TYPE_EOF \
               and _KEYSTORE.get(key, 'part_number') == len(parts):
                LOG.info("upload part complete")
                _KEYSTORE.set(key, 'payload', {})
        except Exception, e:
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            del body, payload
            gc.collect()


class XtrargonUppartComplete(object):
    def __init__(self, fsdbucket, fsdclient):
        self.fsdbucket = fsdbucket
        self.fsdclient = fsdclient

    def run(self, key):
        res = _KEYSTORE.acquire(key, exists = True)
        if res is None:
            LOG.debug("object complete: doesn't exist: %r", key)
            return

        try:
            if _KEYSTORE.get(key, 'size') \
               or _KEYSTORE.get(key, 'last_chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
                return

            upload_id = _KEYSTORE.get(key, 'id')
            size      = _KEYSTORE.get(key, 'size')
            parts     = _KEYSTORE.get(key, 'parts')

            mupload   = {'Parts': []}
            for part in sorted(parts, key = lambda k: k['PartNumber']):
                mupload['Parts'].append(part)

            LOG.info("object part complete: %r, size: %r, parts: %r", key, size, len(parts))

            self.fsdclient.complete_multipart_upload(
                Bucket          = self.fsdbucket,
                Key             = key,
                UploadId        = upload_id,
                MultipartUpload = mupload)

            LOG.info("object part complete transferred: %r, size: %r, parts: %r", key, size, len(parts))

            _KEYSTORE.delete(key)
        except Exception, e:
            LOG.exception(e)
        finally:
            _KEYSTORE.try_release(key)
            gc.collect()


class XtrargonUploadJob(threading.Thread):
    def __init__(self, queue, uppool, upnorm, uppart, upcomp, writef, compress_level = False):
        threading.Thread.__init__(self)

        self._killed        = False

        self.queue          = queue
        self.uppool         = uppool
        self.upnorm         = upnorm
        self.uppart         = uppart
        self.upcomp         = upcomp
        self.writef         = writef
        self.compress_level = compress_level

    def kill(self):
        self._killed = True

    def killed(self):
        return self._killed

    def _set_keystore(self, uid, key):
        if not _KEYSTORE.has_key(key, 'last_chunk_type'):
            (_KEYSTORE.set(key, 'payload', {})
                      .set(key, 'size', 0)
                      .set(key, 'id', None)
                      .set(key, 'part_number', 0)
                      .set(key, 'first_chunk_no', _KEYUID.get(uid, 'chunk_no'))
                      .set(key, 'last_chunk_no', None)
                      .set(key, 'last_chunk_type', None)
                      .set(key, 'is_ibdata', _KEYUID.get(uid, 'is_ibdata'))
                      .set(key, 'writef', _KEYUID.get(uid, 'writef'))
                      .set(key, 'dirname', _KEYUID.get(uid, 'dirname'))
                      .set(key, 'filename', _KEYUID.get(uid, 'filename'))
                      .set(key, 'start_part', None)
                      .set(key, 'wait_part', None)
                      .set(key, 'parts', []))

        size     = _KEYUID.get(uid, 'size')
        if size is None:
            return

        chunk_no = int(_KEYUID.get(uid, 'chunk_no'))

        if _KEYSTORE.get(key, 'wait_part') == chunk_no:
            _KEYSTORE.set(key, 'wait_part', None)

        payload = _KEYUID.get(uid, 'payload')

        if self.compress_level is not False:
            LOG.debug("begin compressing: %r", uid)
            payload = zlib.compress(payload, self.compress_level)
            size    = len(payload)
            LOG.debug("end compressing: %r", uid)

        _KEYSTORE.set(key, 'size', _KEYSTORE.get(key, 'size') + size)
        _KEYSTORE.get(key, 'payload')[chunk_no] = (size, payload)

        payload = None
        _KEYUID.set(uid, 'size', None)
        _KEYUID.set(uid, 'payload', "")

        if _KEYUID.get(uid, 'status') != 'ready':
            return

        _KEYUID.set(uid, 'status', 'deleted')
        _KEYSTORE.set(key, 'last_chunk_type', _KEYUID.get(uid, 'chunk_type'))
        _KEYSTORE.set(key, 'last_chunk_no', chunk_no)

    def run(self):
        while not _KILLED and not self.killed():
            try:
                (uid, key) = self.queue.get(True, 0.1)
            except Queue.Empty:
                continue

            if not _KEYSTORE.acquire(key, exists = True):
                if uid:
                    _KEYUID.delete(uid)
                continue

            if uid \
               and _KEYUID.has_key(uid, 'status') \
               and _KEYUID.expired(uid, OBJECT_MAX_LIFETIME):
                _KEYUID.delete(uid)
                uid = None

            if _KEYSTORE.has_key(key, 'last_chunk_type') \
               and _KEYSTORE.expired(key, OBJECT_MAX_LIFETIME):
                LOG.warning("object expired: %r", key)
                _KEYSTORE.delete(key)
                if uid:
                    _KEYUID.delete(uid)
                continue

            try:
                if uid:
                    self._set_keystore(uid, key)

                size            = _KEYSTORE.get(key, 'size')
                chunk_type      = _KEYSTORE.get(key, 'last_chunk_type')
                first_chunk_no  = _KEYSTORE.get(key, 'first_chunk_no')
                last_chunk_no   = _KEYSTORE.get(key, 'last_chunk_no')
                wait_part       = _KEYSTORE.get(key, 'wait_part')

                if size >= S3_MIN_CHUNK_SIZE \
                   and not _KEYSTORE.get(key, 'writef') \
                   and not _KEYSTORE.get(key, 'id'):
                    _KEYSTORE.set(key, 'id', True)

                if wait_part is not None:
                    if last_chunk_no is not None and wait_part >= last_chunk_no:
                        _KEYSTORE.set(key, 'wait_part', None)
                    else:
                        LOG.debug("wait_part: %r, %r, %r", wait_part, key, last_chunk_no)
                        continue

                if _KEYSTORE.get(key, 'writef'):
                    if size == 0 \
                       and chunk_type != xbstream.XB_CHUNK_TYPE_EOF:
                        continue
                    _KEYSTORE.try_release(key)
                    self.uppool.run(self.writef.run, None, 'writef', *[key])
                elif not _KEYSTORE.get(key, 'id'):
                    if size == 0 \
                       or chunk_type != xbstream.XB_CHUNK_TYPE_EOF:
                        continue
                    _KEYSTORE.try_release(key)
                    self.uppool.run(self.upnorm.run, None, "upnorm", *[key])
                elif size == 0 \
                   and not _KEYSTORE.get(key, 'payload') \
                   and chunk_type == xbstream.XB_CHUNK_TYPE_EOF:
                    _KEYSTORE.try_release(key)
                    self.uppool.run(self.upcomp.run, None, "upcomp", *[key])
                else:
                    if size == 0:
                        continue
                    _KEYSTORE.try_release(key)
                    self.uppool.run(self.uppart.run, None, "uppart", *[key])
            except Exception, e:
                LOG.exception(e)
            finally:
                if uid and _KEYUID.get(uid, 'status') == 'deleted':
                    _KEYUID.delete(uid)
                    uid = None

                if not uid or _KEYUID.get(uid, 'chunk_no') == first_chunk_no:
                    _KEYSTORE.try_release(key)
                    self.queue.put_nowait((uid, key))
                else:
                    _KEYSTORE.try_release(key)
                gc.collect()


class XtrargonJobDispatcher(threading.Thread):
    def __init__(self, uppool, fsdbucket, fsdclient, keys = None, jobs = None, compress_level = False):
        threading.Thread.__init__(self)

        self.uppool         = uppool
        self.upnorm         = XtrargonUploadNormal(fsdbucket, fsdclient)
        self.uppart         = XtrargonUploadPart(fsdbucket, fsdclient)
        self.upcomp         = XtrargonUppartComplete(fsdbucket, fsdclient)
        self.writef         = XtrargonWriteFile()
        self.compress_level = compress_level
        self.keys           = keys or {}
        self.jobs           = jobs or Queue.Queue()

    def _is_ready(self, uid, key):
        if _KEYUID.get(uid, 'chunk_type') != xbstream.XB_CHUNK_TYPE_EOF:
            return

        r = True
        c = _KEYUID.get(uid, 'chunk_no')

        for no in range(c):
            kuid = "%s.%s" % (key, no + 1)
            if not _KEYUID.exists(kuid) \
               or _KEYUID.get(kuid, 'status') == 'processing':
                r = False
                break

        if r:
            _KEYUID.set(uid, 'status', 'ready')
        return r

    def run(self):
        while not _KILLED:
            try:
                uid = _QUPJOB.get(True, 60)
            except Queue.Empty:
                while True:
                    try:
                        job = self.jobs.get(False)
                    except Queue.Empty:
                        continue
                    if len(_KEYSTORE.list()) == 0:
                        job.kill()
                    elif not job.killed():
                        self.jobs.put(job)
                gc.collect()
                continue

            key = _KEYUID.get(uid, 'key')
            self._is_ready(uid, key)

            if key in self.keys and not self.keys[key].killed():
                LOG.debug("%r: added %r", self.keys[key].name, uid)
                self.keys[key].queue.put_nowait((uid, key))
                continue

            cjob = None

            jobs_size = self.jobs.qsize()
            if jobs_size > 0 and len(self.keys) >= JOB_MAX_QUEUE_SIZE:
                for n in range(jobs_size):
                    job = self.jobs.get()
                    if job.killed():
                        continue
                    elif job.queue.qsize() >= JOB_MAX_QUEUE_SIZE:
                        self.jobs.put(job)
                    else:
                        cjob = job
                        break

            if cjob:
                LOG.debug("%r: added %r", cjob.name, uid)
                cjob.queue.put_nowait((uid, key))
            else:
                cjob = XtrargonUploadJob(Queue.Queue(),
                                         self.uppool,
                                         self.upnorm,
                                         self.uppart,
                                         self.upcomp,
                                         self.writef,
                                         self.compress_level)
                cjob.setName("upjob:%d" % (jobs_size + 1))
                LOG.debug("%r: added %r", cjob.name, uid)
                cjob.queue.put_nowait((uid, key))
                cjob.setDaemon(True)
                cjob.start()

            self.keys[key] = cjob
            self.jobs.put(cjob)


class XtrargonStreamTx(threading.Thread):
    def __init__(self, server, path, xbobj, compress_level, max_memory):
        threading.Thread.__init__(self)
        self.server         = server
        self.path           = path
        self.xbobj          = xbobj
        self.compress_level = compress_level
        self.max_memory     = max_memory

    def _mk_realdstpath(self, filepath):
        if not self.server.config.get('paths'):
            return filepath

        xpath = None
        for key in sorted(self.server.config['paths'].keys(), key=len, reverse=True):
            path = os.path.commonprefix([filepath, key])
            if path and path in self.server.config['paths']:
                xpath = path
                break

        if not xpath:
            return filepath

        r   = filepath
        cfg = self.server.config['paths'][path]

        if cfg.get('dst') and filepath.startswith(xpath):
            r = os.path.join(cfg['dst'], filepath[len(xpath):])

        if cfg.get('prefix_server_id'):
            r = os.path.join(os.path.sep, self.server.server_id, r.lstrip(os.path.sep))

        return r

    def run(self):
        if self.compress_level is not False:
            name = "%s.z" % self.xbobj.name
        else:
            name = self.xbobj.name

        is_ibdata = name.startswith('ibdata')
        body      = ""
        key       = self._mk_realdstpath(os.path.join(self.path, name)).lstrip('/')
        uid       = "%s.%s" % (key, self.xbobj.chunk_no)

        try:
            if _SYS.expired('ps', 60):
                _SYS.set('ps', 'info', helpers.ps_info(os.getpid()))

            if not _KEYSTORE.exists(key):
                _KEYSTORE.add(key)

            _KEYUID.add(uid)
            (_KEYUID.set(uid, 'status', 'processing')
                    .set(uid, 'key', key)
                    .set(uid, 'payload', self.xbobj.payload)
                    .set(uid, 'size', self.xbobj.payload_size)
                    .set(uid, 'chunk_no', self.xbobj.chunk_no)
                    .set(uid, 'chunk_type', self.xbobj.chunk_type)
                    .set(uid, 'is_ibdata', is_ibdata)
                    .set(uid, 'writef', is_ibdata or _SYS.get('ps', 'info')['mem_info']['vms'] > self.max_memory)
                    .set(uid, 'dirname', self.path)
                    .set(uid, 'filename', name))
        except Exception, e:
            LOG.exception(e)
        else:
            _KEYUID.set(uid, 'status', 'complete')
            _QUPJOB.put_nowait(uid)
        finally:
            self.xbobj = None
            gc.collect()


class XtrargonStreamHandler(StreamRequestHandler):
    def setup(self):
        StreamRequestHandler.setup(self)
        self.path           = None

        self.max_memory     = bitmath.parse_string_unsafe(
                                 self.server.config.get('max_memory') or DEFAULT_MAX_MEMORY).to_Byte().value

        self.callbacks      = xbstream.XBStreamDefaultCallbacks(read = self.rfile.read,
                                                                tx   = self._tx)

    def _tx(self, xbobj):
        t = XtrargonStreamTx(self.server, self.path, xbobj, self.server.compress_level, self.max_memory)
        t.setDaemon(True)
        t.start()

    def handle(self):

        try:
            r = ""
            while not _KILLED:
                data = self.rfile.read(1)
                if data == "":
                    break
                r += data
                if r[-2:] == "\r\n":
                    r = r[:-2]
                    break

            state = pickle.loads(r)
            if not state or state.get('status') != 'processing':
                LOG.error("invalid xtrargon status")
                self.wfile.write("xtrargon: invalid xtrargon status\r\n")
                return
            self.wfile.write("xtrargon: OK\r\n")

            self.path = state['path']
            xbread    = xbstream.XBStreamRead(self.callbacks, set_payload = True)

            while xbread.is_streaming():
                xbread()
        except Exception, e:
            self.wfile.write("xtrargon: exception %r\r\n" % e)
            LOG.exception(e)


class XtrargonStreamServer(threading.Thread, threading_tcp_server.KillableThreadingTCPServer):
    def __init__(self, config, server_id, compress_level, fsdbucket, fsdclient):
        threading.Thread.__init__(self)
        self.server           = threading_tcp_server.KillableThreadingTCPServer(
                                  config,
                                  (config['listen_addr'],
                                   config['listen_port']),
                                  XtrargonStreamHandler,
                                  name = "stream")
        self.server.config          = config
        self.server.server_id       = server_id
        self.server.compress_level  = compress_level
        self.server.fsdbucket       = fsdbucket
        self.server.fsdclient       = fsdclient

    def run(self):
        LOG.info("stream server will now serve")
        while not self.server.killed():
            try:
                self.server.serve_until_killed()
            except (socket.error, select.error), why:
                if errno.EINTR == why[0]:
                    LOG.debug("interrupted system call")
                elif errno.EBADF == why[0] and self.server.killed():
                    LOG.debug("server close")
                else:
                    raise
        LOG.info("stream server exiting")


class XtrargonStreamPlugin(DWhoPluginBase):
    PLUGIN_NAME = 'stream'

    def safe_init(self):
        self.stream_server  = None
        self.fsdbucket      = None
        self.uppool         = None
        self.compress_level = False

        _SYS.set('ps', 'info', helpers.ps_info(os.getpid()))

        resource.setrlimit(resource.RLIMIT_NOFILE, (RLIMIT_NOFILE, RLIMIT_NOFILE))

        if self.config['plugins']['stream'].get('credentials'):
            cred = helpers.load_yaml_file(self.config['plugins']['stream']['credentials'])
            if not cred:
                raise ValueError("unable to read credentials")

            for k, v in cred['s3'].iteritems():
                os.environ[k.upper()] = v

            if cred['s3'].get('aws_default_bucket'):
                self.fsdbucket = cred['s3']['aws_default_bucket']

        if not self.fsdbucket and os.environ.get('AWS_DEFAULT_BUCKET'):
            self.fsdbucket = os.environ['AWS_DEFAULT_BUCKET']

        if not self.fsdbucket:
            raise ValueError("missing default bucket setting")

        config    = self.config['plugins']['stream']
        cfgclient = None

        if config and 'client' in config:
            cfgclient = botocore.client.Config(**config['client'])

        self.fsdclient = boto3.client('s3', config = cfgclient)

    def at_start(self):
        if 'compress' in self.config['plugins']['stream'] \
           and bool(self.config['plugins']['stream']['compress'].get('enabled')):
            self.compress_level = int(self.config['plugins']['stream']['compress']
                                                 .get('level', DEFAULT_COMPRESS_LEVEL))

        worker_lifetime     = int(self.config['plugins']['stream'].get('worker_lifetime')) or WORKER_LIFETIME

        self.stream_server  = XtrargonStreamServer(self.config['plugins']['stream'],
                                                   self.server_id,
                                                   self.compress_level,
                                                   self.fsdbucket,
                                                   self.fsdclient)
        self.stream_server.setDaemon(True)
        self.stream_server.start()

        self.uppool         = workerpool.WorkerPool(max_workers = self.fsdclient._client_config.max_pool_connections,
                                                    life_time   = worker_lifetime,
                                                    name        = "uppool")

        keys = {}
        jobs = Queue.Queue()

        max_workers         = int(self.config['plugins']['stream'].get('max_workers') or 1)
        if max_workers < 1:
            max_workers = 1

        for n in range(max_workers * 10):
            jdisp               = XtrargonJobDispatcher(self.uppool,
                                                        self.fsdbucket,
                                                        self.fsdclient,
                                                        keys           = keys,
                                                        jobs           = jobs,
                                                        compress_level = self.compress_level)
            jdisp.setDaemon(True)
            jdisp.setName("jdisp:%s" % n)
            jdisp.start()

    def at_stop(self):
        global _KILLED
        _KILLED = True

        if self.stream_server:
            self.stream_server.kill()

        if self.uppool:
            self.uppool.killall(1)


if __name__ != "__main__":
    def _start():
        PLUGINS.register(XtrargonStreamPlugin())
    _start()
