# -*- coding: utf-8 -*-
"""sync plugin"""

__author__  = "Adrien DELLE CAVE"
__license__ = """
    Copyright (C) 2017  fjord-technologies

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA..
"""

import bitmath
import boto3
import botocore
import logging
import pyinotify
import os
import threading

from awscli import clidriver, EnvironmentVariables, utils
from boto3.s3 import transfer as S3Transfer

from dwho.adapters.redis import DWhoAdapterRedis
from dwho.classes.inoplugs import DWhoInoEventPlugBase, INOPLUGS
from sonicprobe import helpers


LOG = logging.getLogger('xtrargon.plugins.sync')

logging.getLogger('awscli').setLevel(logging.WARNING)
logging.getLogger('awscli.clidriver').setLevel(logging.WARNING)
logging.getLogger('boto3').setLevel(logging.WARNING)
logging.getLogger('botocore').setLevel(logging.WARNING)
logging.getLogger('requests').setLevel(logging.WARNING)

_KILLED       = False
XTRARGON_FILE = ".xtrargon"


class XtrargonSyncTxProgress(object):
    def __init__(self, filename):
        self._filename = filename
        self._size = float(os.path.getsize(filename))
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount):
        with self._lock:
            if _KILLED:
                raise KeyboardInterrupt()
            self._seen_so_far += bytes_amount
            if self._size:
                percentage = (self._seen_so_far / self._size) * 100
            else:
                percentage = 100

            if not (int(percentage) % 25):
                LOG.info("XtrargonSyncTxProgress: %r  %s / %s  (%.2f%%)",
                         self._filename,
                         bitmath.Byte(self._seen_so_far).best_prefix().format('{value:.2f} {unit}'),
                         bitmath.Byte(self._size).best_prefix().format('{value:.2f} {unit}'),
                         percentage)


class XtrargonSyncInoPlug(DWhoInoEventPlugBase):
    PLUGIN_NAME = 'sync'

    @staticmethod
    def _create_driver():
        emitter = clidriver.HierarchicalEmitter()
        session = botocore.session.Session(EnvironmentVariables, event_hooks=emitter)
        clidriver._set_user_agent_for_session(session)
        clidriver.load_plugins(session.full_config.get('plugins', {}),
                               event_hooks=emitter)
        return clidriver.CLIDriver(session=session).main

    def safe_init(self):
        self.bucket = None

        if self.config['inotify']['plugins']['sync'].get('credentials'):
            cred = helpers.load_yaml_file(self.config['inotify']['plugins']['sync']['credentials'])
            if not cred:
                raise ValueError("unable to read credentials")

            for k, v in cred['s3'].iteritems():
                os.environ[k.upper()] = v

            if cred['s3'].get('aws_default_bucket'):
                self.bucket = cred['s3']['aws_default_bucket']

        if not self.bucket and os.environ.get('AWS_DEFAULT_BUCKET'):
            self.bucket = os.environ['AWS_DEFAULT_BUCKET']

        if not self.bucket:
            raise ValueError("missing default bucket setting")

        config    = self.config['inotify']['plugins']['sync']
        cfgclient = None
        cfgtx     = None

        if 'client' in config:
            cfgclient = botocore.client.Config(**config['client'])
        if 'transfer' in config:
            cfgtx     = S3Transfer.TransferConfig(**config['transfer'])

        self.client         = boto3.client('s3', config = cfgclient)
        self.transfer       = S3Transfer.S3Transfer(self.client, config = cfgtx)
        self.driver         = self._create_driver()
        self.adapter_redis  = DWhoAdapterRedis(self.config, prefix = 'file_sync')

    def at_stop(self):
        global _KILLED
        _KILLED = True

    def _s3uri(self, path = None):
        uri = "s3://%s" % self.bucket
        if path:
            uri += "/%s" % path.lstrip('/')

        return uri

    def _s3cmd(self, *args):
        LOG.info("_s3cmd: %r", args)
        try:
            return self.driver(('s3',) + args)
        except Exception, e:
            LOG.error(repr(e))

    def _s3api(self, cmd, key = None, **kwargs):
        kwargs['Bucket'] = self.bucket
        if key:
            kwargs['Key'] = key.lstrip('/')

        LOG.info("_s3api: %r: %r", cmd, kwargs)
        try:
            return getattr(self.client, cmd)(**kwargs)
        except Exception, e:
            LOG.error(repr(e))

    def _s3up(self, src, dst = None):
        if not dst:
            dst = src

        LOG.info("_s3up (src: %r, dst: %r, bucket: %r)",
                 src,
                 dst,
                 self.bucket)
        try:
            return self.transfer.upload_file(src,
                                             self.bucket,
                                             dst.lstrip('/'),
                                             callback = XtrargonSyncTxProgress(src))
        except Exception, e:
            LOG.error(repr(e))

    def _s3_touch_xtrargon(self, path):
        return self._s3api('put_object', os.path.join(path, XTRARGON_FILE), **{'Body': ''})

    @staticmethod
    def _mk_sum(filepath):
        if not os.path.exists(filepath):
            return ''

        fstat = os.stat(filepath)

        return '|'.join((str(fstat.st_size),
                         str(fstat.st_mtime)))

    def _mk_filekey(self, filepath):
        return "%s:file:%s" % (self.server_id, filepath)

    def _mk_cookiekey(self, cookie):
        return "%s:cookie:%s" % (self.server_id, cookie)

    def _mk_realdstpath(self, event, filepath):
        if self.cfg_path.path not in self.config['inotify']['paths']:
            return filepath

        r   = filepath
        cfg = self.config['inotify']['paths'][self.cfg_path.path]

        if cfg.get('dst') and filepath.startswith(self.cfg_path.path):
            r = os.path.join(cfg['dst'], filepath[len(self.cfg_path.path):])

        if cfg.get('prefix_server_id'):
            r = os.path.join(os.path.sep, self.server_id, r.lstrip(os.path.sep))

        return r

    def __call__(self, cfg_path, event, filepath):
        self.cfg_path = cfg_path

        if event.dir:
            filepath += '/'

        dstpath = self._mk_realdstpath(event, filepath)
        filekey = self._mk_filekey(dstpath)

        if event.mask & pyinotify.IN_MOVED_FROM != 0:
            cookiekey = self._mk_cookiekey(event.cookie)
            self.adapter_redis.set_key(cookiekey,
                                       dstpath)
            return self

        if event.mask == (pyinotify.IN_CREATE | pyinotify.IN_ISDIR):
            osum = self.adapter_redis.get_key(filekey)
            nsum = self._mk_sum(filepath)
            self._s3_touch_xtrargon(dstpath)
            self.adapter_redis.set_key(filekey, nsum)
        elif event.mask == (pyinotify.IN_MOVED_TO | pyinotify.IN_ISDIR):
            nsum      = self._mk_sum(filepath)
            cookiekey = self._mk_cookiekey(event.cookie)
            pdstpath  = self.adapter_redis.get_key(cookiekey)

            if pdstpath:
                self._s3cmd('mv', self._s3uri(pdstpath), self._s3uri(dstpath), '--recursive')
            self._s3_touch_xtrargon(dstpath)

            self.adapter_redis.del_key((cookiekey, self._mk_filekey(pdstpath)))
            self.adapter_redis.set_key(filekey, nsum)
        elif event.mask == pyinotify.IN_CLOSE_WRITE:
            osum      = self.adapter_redis.get_key(filekey)
            nsum      = self._mk_sum(filepath)
            if not osum or osum != nsum:
                if filepath.endswith('.qp'):
                    t = threading.Thread(target = helpers.gzipfile,
                                         kwargs = {'src': filepath,
                                                   'dst': "%s.gz" % filepath,
                                                   'delete': True})
                    t.setDaemon(True)
                    t.start()
                else:
                    self._s3up(filepath, dstpath)
            self.adapter_redis.set_key(filekey, nsum)
        elif event.mask == pyinotify.IN_MOVED_TO:
            nsum      = self._mk_sum(filepath)
            cookiekey = self._mk_cookiekey(event.cookie)
            pdstpath  = self.adapter_redis.get_key(cookiekey)

            if pdstpath:
                if self._s3cmd('mv', self._s3uri(pdstpath), self._s3uri(dstpath)):
                    self._s3up(filepath, dstpath)
            else:
                self._s3up(filepath, dstpath)

            self.adapter_redis.del_key((cookiekey, self._mk_filekey(pdstpath)))
            self.adapter_redis.set_key(filekey, nsum)

        return self


if __name__ != "__main__":
    def _start():
        INOPLUGS.register(XtrargonSyncInoPlug())
    _start()
